#include <sys/types.h>
#include <sys/stat.h>

#include <net/if.h>
#include <net/pfvar.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <time.h>
#include <inttypes.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <stdbool.h>
#include <err.h>

uint32_t jenkins_hash32(const uint32_t *, size_t, uint32_t);
uint32_t murmur3_aligned_32(const void *, size_t, uint32_t);

#define	STATSIZE	100

int
main(int argc, char *argv[])
{
	struct stat st;
	struct timespec ts1, ts2;
	struct pf_state_key_cmp *key;
	bool jenkins, murmur;
	int fd, nkeys, nkeys2, ch;
	uint32_t col[32768];
	int colst[STATSIZE];

	jenkins = murmur = false;

	while ((ch = getopt(argc, argv, "jm")) != -1)
		switch(ch) {
		case 'm':
			murmur = true;
			break;
		case 'j':
			jenkins = true;
			break;
		default:
			errx(1, "getopt");
		}
	argc -= optind;
	argv += optind;

	if (jenkins == false && murmur == false)
		errx(1, "no hash");
	if (jenkins == true && murmur == true)
		errx(1, "two hashes");
	if (argc == 0)
		errx(1, "no file");

	if (stat(argv[0], &st) != 0)
		err(1, "stat(%s)", argv[0]);

	nkeys = st.st_size / sizeof(*key);
	if (nkeys * sizeof(*key) != st.st_size)
		errx(1, "size %ju not multiple of %ju",
		    (uintmax_t)st.st_size, (uintmax_t)sizeof(*key));

	if ((key = malloc(nkeys * sizeof(*key))) == NULL)
		err(1, "malloc");

	if ((fd = open(argv[0], O_RDONLY)) < 0)
		err(1, "open(%s)", argv[0]);

	if (read(fd, key, nkeys * sizeof(*key)) != nkeys * sizeof(*key))
		err(1, "read");

	bzero(col, sizeof(col));

	clock_gettime(CLOCK_MONOTONIC_PRECISE, &ts1);
	for (int i = 0; i < nkeys; i++) {
		uint32_t h;

		if (jenkins)
			h = jenkins_hash32((uint32_t *)&key[i],
			    sizeof(struct pf_state_key_cmp)/sizeof(uint32_t),
			    2785035390);
		if (murmur)
			h = murmur3_aligned_32((uint32_t *)&key[i],
			    sizeof(struct pf_state_key_cmp)/sizeof(uint32_t),
			    2785035390);

		h &= 32767;

		col[h]++;
	}
	clock_gettime(CLOCK_MONOTONIC_PRECISE, &ts2);

	for (int i = 0; i < 32768; i++) {
		if (col[i] > STATSIZE - 1)
			col[i] = STATSIZE - 1;
		colst[col[i]]++;
	}

	nkeys2 = 0;
	for (int i = 0; i < STATSIZE; i++)
		if (colst[i] > 0) {
			nkeys2 += i * colst[i];
			printf("%d collisions %d hits\n", i, colst[i]);
		}

	printf("Read %d keys, found %d keys in hash\n", nkeys, nkeys2);
	printf("Execution time %lu nsec\n", ts2.tv_nsec - ts1.tv_nsec);

#if 0
	printf("%d %s:%u %s:%u\n", key.proto,
	    inet_ntoa(key.addr[0].v4), ntohs(key.port[0]),
	    inet_ntoa(key.addr[1].v4), ntohs(key.port[0]));
#endif
}
